#!/usr/bin/perl -w
#
#  MODFLOW Object
#  Use this module to read, write and manipulate MODFLOW files
#
#  Copyright (c) 2001  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along 
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#  See also www.fsf.org

package MODFLOW;

use strict;
use Search;
use FORTRAN;
use Geometry;

sub new
{
   my $class = shift;
   my $self = {};
   bless($self,$class);

   my %args = ('X0'  => 0 ,
               'Y0'  => 0 ,
               'Ni'  => undef ,
               'Nj'  => undef ,
               'Nk'  => 1 ,
               'DX'  => undef ,
               'DY'  => undef ,
               'DOUBLE' => 0 ,
               @_);

   $self->{Ni} = $args{Ni};
   $self->{Nj} = $args{Nj};
   $self->{Nk} = $args{Nk};
   $self->{DP} = $args{DOUBLE};
   $self->{VAR} = ();
   %{$self->{IO}} = ();
   defined($args{Ni}) && defined($args{Nj}) && defined($args{Nk}) || die "MODFLOW dimensions not specified\n";
 
   if (!defined($args{DX}))
   {
      $self->{Xm} = [];
   }
   elsif (ref($args{DX}) eq 'ARRAY')
   {
      (@{$args{DX}}==$self->{Nj}) || die "MODFLOW DX count does not match Nj\n";
      $self->{Xm}[0] = $args{X0};
      for (my $j=0;$j<$self->{Nj};$j++)
      {
         $self->{Xm}[$j+1] = $self->{Xm}[$j] + $args{DX}[$j];
      }
   }
   else
   {
      for (my $j=0;$j<=$self->{Nj};$j++)
      {
         $self->{Xm}[$j] = $args{X0}+$j/$self->{Nj}*$args{DX};
      }
   }
   for (my $j=1;$j<@{$self->{Xm}};$j++) #  Also works when DX is undefined
   {
      $self->{X}[$j] = ($self->{Xm}[$j] + $self->{Xm}[$j-1])/2;
      $self->{DX}[$j] = $self->{Xm}[$j] - $self->{Xm}[$j-1];
   }

   if (!defined($args{DY}))
   {
      $self->{Ym} = [];
   }
   elsif (ref($args{DY}) eq 'ARRAY')
   {
      (@{$args{DY}}==$self->{Ni}) || die "MODFLOW DY count does not match Ni\n";
      $self->{Ym}[$self->{Ni}] = $args{Y0};
      for (my $i=$self->{Ni};$i>0;$i--)
      {
         $self->{Ym}[$i-1] = $self->{Ym}[$i] + $args{DY}[$i-1];
      }
   }
   else
   {
      for (my $i=0;$i<=$self->{Ni};$i++)
      {
         $self->{Ym}[$i] = $args{Y0}+($self->{Ni}-$i)/$self->{Ni}*$args{DY};
      }
   }
   for (my $i=1;$i<@{$self->{Ym}};$i++) #  Also works when DY is undefined
   {
      $self->{Y}[$i] = ($self->{Ym}[$i] + $self->{Ym}[$i-1])/2;
      $self->{DY}[$i] = $self->{Ym}[$i-1] - $self->{Ym}[$i];
   }

   $self->{Nij}  = $self->{Ni}*$self->{Nj};
   $self->{Nijk} = $self->{Ni}*$self->{Nj}*$self->{Nk};
   $self->{X0} = $self->{Xm}[0];
   $self->{X1} = $self->{Xm}[$self->{Nj}];
   $self->{Y1} = $self->{Ym}[0];
   $self->{Y0} = $self->{Ym}[$self->{Ni}];

   return $self;
}

#
#  Parse format
#
sub unformat
{
   my ($fmt) = @_;
   $fmt =~ s/ //g;
   (substr($fmt,0,1) eq "(" && substr($fmt,-1,1) eq ")") || die "Invalid format $fmt\n";
   $fmt = substr($fmt,1,-1);
   if ($fmt eq "*")
   {
      return;
   }
   else
   {
      my ($cnt,$wd,$junk) = split(/[IFEGifeg]/,$fmt);
      my ($wid,$dec) = split(/\./,$wd);
      my $typ = 'd';
      defined($dec) && ($dec>0) && ($typ = ($fmt =~ /[EGeg]/) ? 'e' : 'f');
      ($cnt>0 && $wid>0 && !defined($junk)) || die "Format too complex $fmt\n";
      return ($cnt,$wid,$dec,$typ);
   }
}

#
#  Find (x,y) in domain
#
sub find
{
   my $self = shift;
   my ($x,$y) = @_;
   (@_==2) || die "Wrong number of arguments for MODFLOW find";

   # Check inside domain
   my $i = Search::binsearch($y,\@{$self->{Ym}});
   my $j = Search::binsearch($x,\@{$self->{Xm}});
   if (defined($i) && defined($j))
   {
      $i = ($i < $self->{Ni}) ? $i+1 : $self->{Ni};
      $j = ($j < $self->{Nj}) ? $j+1 : $self->{Nj};
      return ($i,$j);
   }
   else
   {
      return;
   }
}

#
#  Map line segment to domain
#
sub line
{
   my $self = shift;
   (@_==4) || die "Wrong number of arguments for MODFLOW line";

   #  Clip to domain
   my ($x0,$y0,$x1,$y1) = Geometry::ClipLine(@_ , $self->{X0},$self->{X1},$self->{Y0},$self->{Y1});

   #  Trap complete miss
   defined($x0) || return;

   #  Starting and ending cells
   my ($i0,$j0) = $self->find($x0,$y0);
   my ($i1,$j1) = $self->find($x1,$y1);

   #  Trap single cell (which includes degenerate lines)
   my %cell;
   my $D = sqrt(($x1-$x0)**2+($y1-$y0)**2);
   if ($i0==$i1 && $j0==$j1)
   {
      $cell{"$i0:$j0"} = $D;
   }
   #  Find all intersections
   else
   {
      my @f = (0,1);
      #  Row crossings
      my $I0 = ($i0<$i1) ? $i0 : $i1;
      my $I1 = ($i0>$i1) ? $i0 : $i1;
      for (my $i=$I0;$i<$I1;$i++)
      {
         push @f , ($self->{Ym}[$i]-$y0)/($y1-$y0);
      }
      #  Column crossings
      my $J0 = ($j0<$j1) ? $j0 : $j1;
      my $J1 = ($j0>$j1) ? $j0 : $j1;
      for (my $j=$J0;$j<$J1;$j++)
      {
         push @f , ($self->{Xm}[$j]-$x0)/($x1-$x0);
      }
      #  Sort f ascending
      @f = sort {$a <=> $b} @f;
      #  Mark cells
      for (my $k=1;$k<@f;$k++)
      {
         my $f = ($f[$k-1]+$f[$k])/2;
         my $x = (1-$f)*$x0 + $f*$x1;
         my $y = (1-$f)*$y0 + $f*$y1;
         my ($i,$j) = $self->find($x,$y);
         $cell{"$i:$j"} += ($f[$k]-$f[$k-1])*$D;
      }
   }
   return %cell;
}

#
#  Map line segment to domain with integration
#
sub lineint
{
   my $self = shift;
   (@_==6) || die "Wrong number of arguments for MODFLOW lineint";

   #  Clip to domain
   my ($x0,$y0,$z0,$x1,$y1,$z1) = Geometry::ClipLineZ(@_ , $self->{X0},$self->{X1},$self->{Y0},$self->{Y1});

   #  Trap complete miss
   defined($x0) || return;

   #  Starting and ending cells
   my ($i0,$j0) = $self->find($x0,$y0);
   my ($i1,$j1) = $self->find($x1,$y1);

   #  Trap single cell (which includes degenerate lines)
   my %cell;
   my $D = sqrt(($x1-$x0)**2+($y1-$y0)**2);
   if ($i0==$i1 && $j0==$j1)
   {
      $cell{"$i0:$j0"}{D} += $D;
      $cell{"$i0:$j0"}{Z} += $D*($z0+$z1)/2;
   }
   #  Find all intersections
   else
   {
      my @f = (0,1);
      #  Row crossings
      my $I0 = ($i0<$i1) ? $i0 : $i1;
      my $I1 = ($i0>$i1) ? $i0 : $i1;
      for (my $i=$I0;$i<$I1;$i++)
      {
         push @f , ($self->{Ym}[$i]-$y0)/($y1-$y0);
      }
      #  Column crossings
      my $J0 = ($j0<$j1) ? $j0 : $j1;
      my $J1 = ($j0>$j1) ? $j0 : $j1;
      for (my $j=$J0;$j<$J1;$j++)
      {
         push @f , ($self->{Xm}[$j]-$x0)/($x1-$x0);
      }
      #  Sort f ascending
      @f = sort {$a <=> $b} @f;
      #  Mark cells
      for (my $k=1;$k<@f;$k++)
      {
         my $f = ($f[$k-1]+$f[$k])/2;
         my $x = (1-$f)*$x0 + $f*$x1;
         my $y = (1-$f)*$y0 + $f*$y1;
         my $z = (1-$f)*$z0 + $f*$z1;
         my $d = ($f[$k]-$f[$k-1])*$D;
         my ($i,$j) = $self->find($x,$y);
         $cell{"$i:$j"}{D} += $d;
         $cell{"$i:$j"}{Z} += $z*$d;
      }
   }
   return %cell;
}

#
#  Map polyline to domain
#
sub polyline
{
   my $self = shift;
   my ($X,$Y) = @_;
   (@_==2) || die "Wrong number of arguments for MODFLOW polyline";
   my %Cell;
   for (my $i=0;$i<@$X;$i++)
   {
      for (my $j=1;$j<@{$X->[$i]};$j++)
      {
         my %cell = $self->line($X->[$i][$j-1],$Y->[$i][$j-1],$X->[$i][$j],$Y->[$i][$j]);
         foreach my $ij (keys %cell)
         {
            $Cell{$ij} += $cell{$ij};
         }
      }
   }
   return %Cell;
}

#
#  Map polygon to domain
#
sub polygon
{
   my $self = shift;
   my ($X,$Y) = @_;
   (@_==2) || die "Wrong number of arguments for MODFLOW polygon";
   #  Determine bounding box
   my ($Xmin,$Xmax,$Ymin,$Ymax);
   for (my $i=0;$i<@$X;$i++)
   {
      for (my $j=0;$j<@{$X->[$i]};$j++)
      {
         (!defined($Xmin) || $Xmin > $X->[$i][$j]) && ($Xmin = $X->[$i][$j]);
         (!defined($Xmax) || $Xmax < $X->[$i][$j]) && ($Xmax = $X->[$i][$j]);
         (!defined($Ymin) || $Ymin > $Y->[$i][$j]) && ($Ymin = $Y->[$i][$j]);
         (!defined($Ymax) || $Ymax < $Y->[$i][$j]) && ($Ymax = $Y->[$i][$j]);
      }
   }
   #  Check for a complete miss
   ($Xmax<$self->{X0} || $Xmin>$self->{X1} || $Ymax<$self->{Y0} || $Ymin>$self->{Y1}) && return;
   #  Limit to domain
   ($Xmin<$self->{X0}) && ($Xmin = $self->{X0});
   ($Xmax>$self->{X1}) && ($Xmax = $self->{X1});
   ($Ymin<$self->{Y0}) && ($Ymin = $self->{Y0});
   ($Ymax>$self->{Y1}) && ($Ymax = $self->{Y1});
   #  Cell limits
   my ($i0,$j0) = $self->find($Xmin,$Ymax);
   my ($i1,$j1) = $self->find($Xmax,$Ymin);
   #  Loop over cells
   my %cell;
   foreach my $i ($i0 .. $i1)
   {
      foreach my $j ($j0 .. $j1)
      {
         my $A = Geometry::PolyArea(Geometry::ClipPoly($X,$Y,$self->{Xm}[$j-1],$self->{Xm}[$j],$self->{Ym}[$i],$self->{Ym}[$i-1]));
         ($A>0) && ($cell{"$i:$j"} += $A);
      }
   }
   return %cell;
}

#
#  Return (x,y) for cell
#
sub xy
{
   my $self = shift;
   my ($i,$j) = @_;
   (@_==2) || die "Wrong number of arguments for MODFLOW xy";

   # Check inside domain
   (1 <= $i && $i <= $self->{Ni}) || die "Row out of range $i\n";
   (1 <= $j && $j <= $self->{Nj}) || die "Column out of range $j\n";
   return ($self->{X}[$j],$self->{Y}[$i]);
}

#
#  Return area for cell
#
sub area
{
   my $self = shift;
   my ($i,$j) = @_;
   (@_==2) || die "Wrong number of arguments for MODFLOW area";

   # Check inside domain
   (1 <= $i && $i <= $self->{Ni}) || die "Row out of range $i\n";
   (1 <= $j && $j <= $self->{Nj}) || die "Column out of range $j\n";
   return $self->{DX}[$j]*$self->{DY}[$i];
}

#
#  Trim values outside range
#
sub trim
{
   my $self = shift;
   my ($var,$min,$max) = @_;
   (@_==2 || @_==3) || die "Wrong number of arguments for MODFLOW trim";
   if (defined($min) && defined($max))
   {
      for (my $l=0;$l<@$var;$l++)
      {
         ($min <= $var->[$l] && $var->[$l] <= $max) || ($var->[$l] = undef);
      }
   }
   elsif (defined($min))
   {
      for (my $l=0;$l<@$var;$l++)
      {
         ($min <= $var->[$l]) || ($var->[$l] = undef);
      }
   }
   elsif (defined($max))
   {
      for (my $l=0;$l<@$var;$l++)
      {
         ($var->[$l] <= $max) || ($var->[$l] = undef);
      }
   }
   if (defined(wantarray()))
   {
      my $n=0;
      for (my $l=0;$l<@$var;$l++)
      {
         defined($var->[$l]) && ($n++);
      }
      return $n;
   }
}

#
#  Get dimensions
#
sub dim
{
   my $self = shift;
   return ($self->{Ni},$self->{Nj},$self->{Nk});
}

#
#  Get cell offset
#
sub offset
{
   my $self = shift;
   my ($i,$j,$k) = @_;
   (@_==2 || @_==3) || die "Wrong number of arguments for MODFLOW offset";

   defined($i) && defined($j) || return;

   (1 <= $i && $i <= $self->{Ni}) || die "MODFLOW->offset: Row $i out of range $k,$i,$j\n";
   (1 <= $j && $j <= $self->{Nj}) || die "MODFLOW->offset: Column $j out of range $k,$i,$j\n";
   my $off = $self->{Nj}*($i-1) + $j-1;
   if (defined($k))
   {
      (1 <= $k && $k <= $self->{Nk}) || die "MODFLOW->offset: Layer $k out of range $k,$i,$j\n";
      $off += $self->{Nij}*($k-1);
   }
   return $off;
}

#
#  Get cell kij
#
sub kij
{
   my $self = shift;
   my ($off) = @_;
   (@_==1) || die "Wrong number of arguments for MODFLOW kij";

   my $k = int($off/$self->{Nij})+1;
   $off -= ($k-1)*$self->{Nij};
   my $i = int($off/$self->{Nj})+1;
   $off -= ($i-1)*$self->{Nj};
   my $j = $off+1;
   return ($k,$i,$j);
}

#
#  Get cell value
#
sub cell
{
   my $self = shift;
   my ($var,$i,$j,$k) = @_;
   (@_==3 || @_==4) || die "Wrong number of arguments for MODFLOW cell";
   (@$var==defined($k) ? $self->{Nijk} : $self->{Nij}) || die "Incorrect variable size\n";

   my $off = $self->offset($i,$j,$k);
   return defined($off) ? $var->[$off] : undef;
}

#
#  Get cell center value
#
sub center
{
   my $self = shift;
   my ($var,$x,$y,$k) = @_;
   (@_==3 || @_==4) || die "Wrong number of arguments for MODFLOW center";
   my ($i,$j) = $self->find($x,$y);
   return $self->cell($var,$i,$j,$k);
}

#
#  Get interpolated value
#
sub interpolate
{
   my $self = shift;
   my ($var,$x,$y,$k) = @_;
   (@_==3 || @_==4) || die "Wrong number of arguments for MODFLOW interpolate";
   my ($i0,$j0) = $self->find($x,$y);
   my $v00 = $self->cell($var,$i0,$j0,$k);
   my $val;
   if (defined($v00))
   {
      my $j1 = ($x < $self->{X}[$j0]) ?  (($j0==1) ? 1 : $j0-1)  :  (($j0 < $self->{Nj}) ? $j0+1 : $self->{Nj});
      my $i1 = ($y > $self->{Y}[$i0]) ?  (($i0==1) ? 1 : $i0-1)  :  (($i0 < $self->{Ni}) ? $i0+1 : $self->{Ni});
      my $v01 = $self->cell($var,$i0,$j1,$k);
      my $v10 = $self->cell($var,$i1,$j0,$k);
      my $v11 = $self->cell($var,$i1,$j1,$k);
      my $f1 = ($i0==$i1) ? 0 : ($y - $self->{Y}[$i0])/($self->{Y}[$i1] - $self->{Y}[$i0]);
      my $g1 = ($j0==$j1) ? 0 : ($x - $self->{X}[$j0])/($self->{X}[$j1] - $self->{X}[$j0]);
      my $f0 = 1-$f1;
      my $g0 = 1-$g1;
(0 <= $f1 && $f1 <= 1 && 0 <= $g1 && $g1 <= 1) || die "Crap $x,$y -> $f1 $g1\n";
      if (defined($v10) && defined($v01) && defined($v11))
      {
         $val = $f0*$g0*$v00 + $f1*$g0*$v10 + $f0*$g1*$v01 + $f1*$g1*$v11;
      }
      elsif (defined($v10) && defined($v11))
      {
         $val = (1-$f0-$g1)*$v10 + $f0*$v00 + $g1*$v11;
      }
      elsif (defined($v01) && defined($v11))
      {
         $val = (1-$f1-$g0)*$v01 + $f1*$v11 + $g0*$v00;
      }
      elsif (defined($v10) && defined($v01))
      {
         $val = (1-$f1-$g1)*$v00 + $f1*$v10 + $g1*$v01;
      }
      elsif (defined($v10))
      {
         $val = $f0*$v00 + $f1*$v10;
      }
      elsif (defined($v01))
      {
         $val = $g0*$v00 + $g1*$v01;
      }
      else
      {
         $val = $v00;
      }
   }
   return $val;
}

#
#  Assign array
#
sub assign
{
   my $self = shift;
   my ($var,$ref) = @_;
   (@_==2) || die "Wrong number of arguments for MODFLOW assign";
   $self->{VAR}{$var} = $ref;
}

#
#  Get variable address
#
sub var
{
   my $self = shift;
   my ($var) = @_;
   (@_==1) || die "Wrong number of arguments for MODFLOW ref";
   return $self->{VAR}{$var};
}

#
#  Get variable value
#
sub value
{
   my $self = shift;
   my ($var,$i,$j,$k) = @_;
   (@_==3 || @_==4) || die "Wrong number of arguments for MODFLOW value";
   ($self->{VAR}{$var}) || die "Variable $var not defined\n";

   return $self->cell($self->{VAR}{$var},$i,$j,$k);
}

#
#  Open a MODFLOW style file for reading or writing
#
sub open
{
   my $self = shift;
   my ($file,$type,$opt,$msg) = @_;
   (2<=@_ && @_<=4 ) || die "Wrong number of arguments for MODFLOW open";
   $msg = ($msg) ? " $msg" : '';
   $self->{WR} = ($file =~ /^ *>/);
   my $fd;
   if ($type eq 'CBC' || $type eq 'LAY')
   {
      my $lahey = $opt;
      $fd = FORTRAN->open($file,$lahey);
      defined($fd) || return;
   }
   elsif ($type eq 'TXT')
   {
      defined($opt) && warn "Opt=$opt ignored for MODFLOW->open TXT\n";
      $fd = new FileHandle($file); 
      defined($fd) || return;
   }
   elsif ($type eq 'CELL')
   {
      $fd = new FileHandle($file); 
      defined($fd) || return;
      #  Write max count line
      if ($self->{WR})
      {
         defined($opt) || die "No maximum count for MODFLOW->open CELL\n";
         printf $fd "%10d$msg\n" , $opt;
      }
      #  Read max count line
      else
      {
         <$fd>;
      }
   }
   else
   {
      die "Invalid MODFLOW->open type $type\n";
   }
   my $k=0;
   while (exists($self->{IO}{$k}))
   {
      $k++;
   }
   %{$self->{IO}{$k}} = (FD=>$fd,TYPE=>$type,NAME=>$file);
   return $k;
}

#
#  Backspace a MODFLOW style file (2 records)
#
sub backspace
{
   my $self = shift;
   my ($k,$n) = @_;
   (@_==1 || @_==2) || die "Wrong number of arguments for MODFLOW backspace";
   defined($k) || die "MODFLOW backspace fid is not defined\n";
   exists($self->{IO}{$k}) || die "Invalid MODFLOW backspace fid $k\n";
   ($self->{IO}{$k}{TYPE} ne 'TXT' || $self->{IO}{$k}{TYPE} eq 'CELL') || die "Cannot MODFLOW backspace a text file\n";
   $n = defined($n) ? $n : 1;
   return $self->{IO}{$k}{FD}->backspace(2*$n);
}

#
#  Skip a MODFLOW style file (2 binary records, 1 text line)
#
sub skip
{
   my $self = shift;
   my ($k,$n) = @_;
   (@_==1 || @_==2) || die "Wrong number of arguments for MODFLOW skip";
   defined($k) || die "MODFLOW skip fid is not defined\n";
   exists($self->{IO}{$k}) || die "Invalid MODFLOW skip fid $k\n";
   $n = defined($n) ? $n : 1;
   if ($self->{IO}{$k}{TYPE} eq 'TXT' || $self->{IO}{$k}{TYPE} eq 'CELL')
   {
      my $fd = $self->{IO}{$k}{FD};
      for (my $i=0;$i<$n;$i++)
      {
         my $line = <$fd>;
         defined($line) || return;
      }
      return 1;
   }
   else
   {
      return $self->{IO}{$k}{FD}->skip(2*$n);
   }
}


#
#  Read and trim substr
#
sub substrim
{
   my ($line,$ofs,$wid) = @_;
   my $str = substr($line,$ofs,$wid);
   $str =~ s/^  *//;  $str =~ s/  *$//;
   return $str;
}

#
#  Read a line from MODFLOW style file until search string is matched
#  Returns undef if no match is found
#
sub findline
{
   my $self = shift;
   my $k    = shift;
   defined($k) || die "MODFLOW findline fid is not defined\n";
   exists($self->{IO}{$k}) || die "Invalid MODFLOW findline fid $k\n";
   ($self->{IO}{$k}{TYPE} eq 'TXT') || die "findline permitted only on TXT files\n";
   my $fd = $self->{IO}{$k}{FD};
   my $pat = shift;
   defined($pat) && length($pat)>0 || die "Invalid pattern in findline\n";
   while (my $line = <$fd>)
   {
      ($line =~ /$pat/) && return $line;
   }
   return;
}

#
#  Read a line from MODFLOW style file
#
sub readline
{
   my $self = shift;
   my $k    = shift;
   defined($k) || die "MODFLOW read fid is not defined\n";
   exists($self->{IO}{$k}) || die "Invalid MODFLOW read fid $k\n";
   ($self->{IO}{$k}{TYPE} eq 'TXT') || die "readline permitted only on TXT files\n";
   my $fd = $self->{IO}{$k}{FD};
   my $line = <$fd>;
   return $line;
}

#
#  Write a line to MODFLOW style file
#
sub writeline
{
   my $self = shift;
   my $k    = shift;
   defined($k) || die "MODFLOW write fid is not defined\n";
   exists($self->{IO}{$k}) || die "Invalid MODFLOW write fid $k\n";
   ($self->{IO}{$k}{TYPE} eq 'TXT') || die "writeline permitted only on TXT files\n";
   my $fd = $self->{IO}{$k}{FD};
   foreach my $line (@_)
   {
      print $fd "$line\n";
   }
}

#
#  Read a MODFLOW style file
#
sub read
{
   my $self = shift;
   my $k    = shift;
   defined($k) || die "MODFLOW read fid is not defined\n";
   exists($self->{IO}{$k}) || die "Invalid MODFLOW read fid $k\n";
   my %file = %{$self->{IO}{$k}};
   # CBC records are all in one
   if ($file{TYPE} eq 'CBC')
   {
      (@_==0) || die "Wrong number of arguments for MODFLOW read (CBC)";
      my $buffer = $file{FD}->read();
      my ($kstp,$kper,$var,$Nj,$Ni,$Nk) = unpack("L2A16L3",$buffer);
      $var =~ s/^  *//;  $var =~ s/  *$//;
      ($Ni == $self->{Ni} && $Nj == $self->{Nj} && $Nk == $self->{Nk}) || die "Dimension mismatch reading $file{NAME}\n";
      $buffer = $file{FD}->read();
      my @var  = unpack(($self->{DP} ? "d*" : "f*"),$buffer);
      (@var == $self->{Nijk}) || die "Data length error $file{NAME} : $var\n";
      return ($var,$kper,$kstp,undef,@var);
   }
   # LAY records require Nk reads
   elsif ($file{TYPE} eq 'LAY')
   {
      (@_==0) || die "Wrong number of arguments for MODFLOW read (LAY)";
      my ($name,$per,$step,$time,@var);
      for (my $k=1;$k<=$self->{Nk};$k++)
      {
         my $buffer = $file{FD}->read();
         unless (defined($buffer))
         {
            ($k==1) && return;
            die "Unexpected EOF reading $file{NAME} layer $k\n";
         }
         my ($kstp,$kper,$tper,$totim,$var,$Nj,$Ni,$K) = unpack(($self->{DP} ? "L2d2A16L3" : "L2f2A16L3"),$buffer);
         $var =~ s/^  *//;  $var =~ s/  *$//;
         ($Ni == $self->{Ni} && $Nj == $self->{Nj}) || die "Dimension ($Ni,$Nj) mismatch reading $file{NAME} layer $k\n";
         ($k == $K) || die "Unexpected layer $K reading $file{NAME} layer $k\n";
         if ($k==1)
         {
            $name = $var;
            $per  = $kper;
            $step = $kstp;
            $time = $totim;
         }
         else
         {
            ($name eq $var)   || die "Unexpected variable $var reading $file{NAME} $name layer $k\n";
            ($per  == $kper)  || die "Unexpected period $kper reading $file{NAME} $name $per layer $k\n";
            ($step == $kstp)  || die "Unexpected step $kstp reading $file{NAME} $name $step layer $k\n";
            ($time == $totim) || die "Unexpected time $totim reading $file{NAME} $name $time layer $k\n";
         }
         $buffer = $file{FD}->read();
         push @var , unpack(($self->{DP} ? "d*" : "f*"),$buffer);
      }
      (@var == $self->{Nijk}) || die "Data length error $file{NAME}\n";
      return ($name,$per,$step,$time,@var);
   }
   elsif ($file{TYPE} eq 'TXT')
   {
      my $fd = $file{FD};
      my $fmt = shift @_;
      my $mul = shift @_;
      #  U2DREL line
      if (!defined($fmt))
      {
         my $line = <$fd>;
         $mul = substrim($line,10,10);
         $fmt = substrim($line,20,20);
      }
      ($mul) || ($mul=1);
      (@_==0) || die "Wrong number of arguments for MODFLOW read (TXT)";

      #  Parse format
      my ($cnt,$wid) = unformat($fmt);

      #  Read data
      my @var;
      if ($cnt)
      {
         for (my $i=0;$i<$self->{Ni};$i++)
         {
            my $line;
            for (my $j=0;$j<$self->{Nj};$j++)
            {
               my $k = $j%$cnt;
               ($k) || ($line = <$fd>);
               my $fld = substrim($line,$k*$wid,$wid);
               ($fld ne "") || ($fld=0);
               push @var , $mul*$fld;
            }
         }
      }
      else
      {
         for (my $i=0;$i<$self->{Ni};$i++)
         {
            my @line;
            for (my $j=0;$j<$self->{Nj};$j++)
            {
               if (@line==0)
               {
                  my $line = <$fd>;
                  @line = split(" ",$line);
               }
               push @var , shift @line;
            }
         }
      }
      (@var==$self->{Nij}) || die "Read wrong item count\n";
      return (undef,undef,undef,undef,@var);
   }
   elsif ($file{TYPE} eq 'CELL')
   {
      my $fd = $file{FD};
      my $col = 1;
      if (@_>0)
      {
         $col = (shift @_)-1;
      }
      my $free = shift @_;
      (@_==0) || die "Wrong number of arguments for MODFLOW read (CELL)";

      #  Read count
      my $line = <$fd>;
      ($line) || return;
      chomp $line;
      my ($n) = split(" ",$line);
      #  Read data
      my @var;
      for (my $off=0;$off<$self->{Nijk};$off++)
      {
         $var[$off] = 0;
      }
      for (my $l=0;$l<$n;$l++)
      {
         $line = <$fd>;
         chomp $line;
         if ($free)
         {
            my ($k,$i,$j,@q) = split(" ",$line);
            my $off = $self->offset($i,$j,$k);
            $var[$off] += $q[$col];
         }
         else
         {
            my $k = substr($line, 0,10);
            my $i = substr($line,10,10);
            my $j = substr($line,20,10);
            my $off = $self->offset($i,$j,$k);
            $var[$off] += substr($line,10*$col+30,10);
         }
      }
      return (undef,undef,undef,undef,@var);
   }
   else
   {
      die "Invalid MODFLOW->read type $file{TYPE}\n";
   }
}

#
#  Zero Array
#
sub set
{
   my $self = shift;
   my ($k,$val) = @_;
   (@_==2) || die "Wrong number of arguments for MODFLOW set";
   ($k>0) || die "Incorrect number of MODFLOW set layers";

   return ($val) x ($k*$self->{Nij});
}

#
#  Write a MODFLOW style file
#
sub write
{
   my $self = shift;
   my $k    = shift;
   my $var  = shift;
   defined($k) || die "MODFLOW->write fid is not defined\n";
   defined($var) && ref($var) || die "Invalid variable in MODFLOW->write\n";
   exists($self->{IO}{$k}) || die "Invalid MODFLOW->write fid $k\n";
   my %file = %{$self->{IO}{$k}};
   # CBC records are all in one
   if ($file{TYPE} eq 'CBC')
   {
      (@_==3) || die "Wrong number of arguments for MODFLOW->write (CBC)";
      my ($kstp,$kper,$name) = @_;
      (@$var==$self->{Nijk}) || die "Incorrect Variable dimension in MODFLOW->write CBC\n";
      $file{FD}->write(pack("L2A16L3",$kstp,$kper,$name,$self->{Nj},$self->{Ni},$self->{Nk}));
      $file{FD}->write(pack("f*",@$var));
   }
   # LAY records require Nk writes
   elsif ($file{TYPE} eq 'LAY')
   {
      (@_==5) || die "Wrong number of arguments for MODFLOW->write (LAY)";
      my ($kstp,$kper,$tper,$totim,$name) = @_;
      my $Nk = scalar(@$var)/$self->{Nij};
      ($Nk==1 || $Nk==$self->{Nk}) || die "Incorrect Variable dimension in MODFLOW->write LAY\n";
      my @tmp = @$var;
      for (my $k=1;$k<=$Nk;$k++)
      {
         $file{FD}->write(pack("L2f2A16L3",$kstp,$kper,$tper,$totim,$name,$self->{Nj},$self->{Ni},$k));
         $file{FD}->write(pack("f*",splice @tmp , 0 , $self->{Nij}));
      }
   }
   elsif ($file{TYPE} eq 'TXT')
   {
      my $fd = $file{FD};
      #  Parse format
      (@_==3 || @_==4) || die "Wrong number of arguments for MODFLOW->write (TXT)";
      my $n = @$var;
      (@$var==$self->{Nij}) || die "Incorrect Variable dimension in MODFLOW->write TXT $n not $self->{Nij}\n";
      my ($unit,$mul,$fmt,$msg) = @_;
      defined($msg) || ($msg = '');
      #  When $Mul is not defined omit the U2DREL record
      unless (defined($mul))
      {
      }
      #  Integer unit implies traditional U2DREL
      elsif ($unit =~ /\s*\d+\s*/)
      {
         printf $fd "%10d%10f%20s%10d$msg\n" , $unit , $mul , $fmt , -1;
      }
      #  MF2K U2DREL
      else
      {
         print $fd "$unit $mul $fmt -1 $msg\n";
      }

      #  If unit != 0, write data
      if ($unit)
      {
         #  Parse format
         my ($cnt,$wid,$dec,$typ) = unformat($fmt);
         $fmt = defined($dec) ? "%$wid.$dec$typ" : "%$wid$typ";

         #  Write data
         my $off=0;
         for (my $i=0;$i<$self->{Ni};$i++)
         {
            for (my $j=0;$j<$self->{Nj};$j++)
            {
               printf $fd $fmt , $var->[$off++];
               (($j+1) % $cnt) || (print $fd "\n");
            }
            ($self->{Nj} % $cnt) && (print $fd "\n");
         }
      }
   }
   elsif ($file{TYPE} eq 'CELL')
   {
      my $fd = $file{FD};
      my $msg = shift @_;
      $msg = defined($msg) ? "   $msg" : '';
      (@_==0) || die "Wrong number of arguments for MODFLOW->write (CELL)";
      my $Nk = scalar(@$var)/$self->{Nij};
      ($Nk==1 || $Nk==$self->{Nk}) || die "Incorrect Variable dimension in MODFLOW->write CELL\n";

      #  Get count
      my $n = 0;
      for (my $off=0;$off<$self->{Nijk};$off++)
      {
         $var->[$off] && ($n++);
      }
      #  Write count
      printf $fd "%10d$msg\n" , $n;
      #  Write data
      my $off=0;
      for (my $k=1;$k<=$Nk;$k++)
      {
         for (my $i=1;$i<=$self->{Ni};$i++)
         {
            for (my $j=1;$j<=$self->{Nj};$j++)
            {
               if ($var->[$off])
               {
                  my $val = sprintf '%9.1f'  , $var->[$off];
                  (length($val)>9) && ($val = sprintf '%9.2e'  , $var->[$off]);
                  (length($val)>9) && die "Field overflow $val\n";
                  printf $fd "%10d%10d%10d%10s\n" , $k , $i , $j , $val;
               }
               $off++;
            }
         }
      }
   }
   else
   {
      die "Invalid MODFLOW->write type $file{TYPE}\n";
   }
}

#
#  Close a MODFLOW style file
#
sub close
{
   my $self = shift;
   my ($k) = @_;
   (@_==1) || die "Wrong number of arguments for MODFLOW close";
   defined($k) || die "MODFLOW close fid is not defined\n";
   exists($self->{IO}{$k}) || die "Invalid MODFLOW close fid $k\n";
   if ($self->{IO}{$k}{TYPE} eq 'TXT' || $self->{IO}{$k}{TYPE} eq 'CELL')
   {
      close($self->{IO}{$k}{FD});
   }
   else
   {
      $self->{IO}{$k}{FD}->close();
   }
   delete($self->{IO}{$k});
}

#
#  Open, read and close a MODFLOW file
#
sub readfile
{
   my $self = shift;
   my $fd = $self->open(@_);
   defined($fd) || return;
   my @res = $self->read($fd);
   $self->close($fd);
   return @res;
}

1;
